/*
        TextArea1.c 1.1 for X11R6 & GNU C Compiler

        Copyright (C) 1998, 1999 Syllabub
        Maroontress Fast Software.
*/

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xatom.h>
#include <X11/Xresource.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#ifdef HAS_WCHAR_H
#include <wchar.h>
#endif
#ifdef HAS_WCTYPE_H
#include <wctype.h> /* widec.h$B$,$5$i$K8F$P$l$k(B */
#endif

#include "Resource.h"
#include "VScrollbar.h"
#include "WCString.h"
#include "Text.h"
#include "TextArea.h"
#include "TextArea3.h"
#include "TextArea2.h"
#include "TextArea1.h"

#ifndef max
#define max(a, b) (((a) < (b)) ? (b) : (a))
#endif
#ifndef min
#define min(a, b) (((a) > (b)) ? (b) : (a))
#endif

#ifdef HAS_ISW_FUNCS
#define ispunct_w iswpunct
#define isspace_w iswspace
#else
#define ispunct_w ispunct
#define isspace_w isspace
#endif

void
GetSelectedRegion(SylTextArea *txt, SylLocation *left, SylLocation *right)
{
    if (txt->started.y < txt->current.y) {
	*left = txt->started;
	*right = txt->current;
    }
    else if (txt->started.y > txt->current.y) {
	*left = txt->current;
	*right = txt->started;
    }
    else if (txt->started.x < txt->current.x) {
	*left = txt->started;
	*right = txt->current;
    }
    else {
	*left = txt->current;
	*right = txt->started;
    }
}

unsigned int
SizeOfRegion(SylTextArea *txt, SylLocation *left, SylLocation *right)
{
    int n, m, size;

    if (left->y == right->y)
	return (right->x - left->x);

    size = GetLengthAtLine(txt->tbs, left->y) - left->x;
    if (IsEndOfLineAtLine(txt->tbs, left->y))
	++size;
    for (n = left->y + 1, m = right->y; n < m; ++n) {
	size += GetLengthAtLine(txt->tbs, n);
	if (IsEndOfLineAtLine(txt->tbs, n))
	    ++size;
    }
    size += right->x;
    return (size);
}

void
CopyRegionToWCString(SylTextArea *txt, SylLocation *left, SylLocation *right,
		     wchar_t *wcs)
{
    int y, n, m, len;
    SylTextBlock *tb;
    wchar_t *str;

    if (left->y == right->y) {
	tb = SeekSylTextBlockAtLine(txt->tbs, left->y);
	n = left->y - txt->tbs->current_line;
	str = CreateWCStringFromSylText(tb->body[n], left->x, right->x);
	wstrcpy(wcs, str);
	free(str);
	return;
    }

    tb = SeekSylTextBlockAtLine(txt->tbs, left->y);
    n = left->y - txt->tbs->current_line;
    len = LengthOfSylText(tb->body[n]);
    str = CreateWCStringFromSylText(tb->body[n], left->x, len);
    wstrcpy(wcs, str);
    free(str);
    wcs += len - left->x;
    if (EndOfLineSylText(tb->body[n]))
	*wcs++ = '\n';
    for (y = left->y + 1, m = right->y; y < m; ++y) {
	tb = SeekSylTextBlockAtLine(txt->tbs, y);
	n = y - txt->tbs->current_line;
	len = LengthOfSylText(tb->body[n]);
	str = CreateWCStringFromSylText(tb->body[n], 0, len);
	wstrcpy(wcs, str);
	free(str);
	wcs += len;
	if (EndOfLineSylText(tb->body[n]))
	    *wcs++ = '\n';
    }
    tb = SeekSylTextBlockAtLine(txt->tbs, right->y);
    n = right->y - txt->tbs->current_line;
    len = LengthOfSylText(tb->body[n]);
    str = CreateWCStringFromSylText(tb->body[n], 0, right->x);
    wstrcpy(wcs, str);
    free(str);
}

void
SetSelectionOwner(SylTextArea *txt, Time tm)
{
#if 0
    SylLocation left, right;
    unsigned int len, bytes;
    wchar_t *wcs;
    char *mbs;
#endif

    XSetSelectionOwner(txt->disp, XA_PRIMARY, txt->window, tm);
    if (XGetSelectionOwner(txt->disp, XA_PRIMARY) != txt->window)
	txt->selected = False;
    txt->stamp = tm;
#if 0    
    /*
      XXX $BM%@hEY$N9b$$(BTODO:
      $B0J2<$N%+%C%H%P%C%U%!99?7$K4X$9$k=hM}$O!"4X?t(BNiceSylTextArea()$B$+$i(B
      $B$+$i8F$S=P$5$l$kJ}<0$KJQ99$7$?$[$&$,$h$$$@$m$&!#(Btxt->cutbuffer_ready
      == False$B$G(Btxt->selected == True$B$J$i$P0J2<$N=hM}$r9T$&$h$&$K!#:#$N(B
      $B$H$3$m$O%+%C%H%P%C%U%!$KBP$9$k=hM}$OL5;k$9$k$3$H$K$7$F$*$/!#(B
    */
    GetSelectedRegion(txt, &left, &right);
    len = SizeOfRegion(txt, &left, &right);
    /*
      CopyRegionToWCString()$B$O:G8e$K%L%kJ8;z$rIU2C$9$k$N$G!"(B1$BJ8;zJ,B?$/(B
      $B3NJ]$7$F$*$/!#(B
    */
    if ((wcs = (wchar_t *)malloc(sizeof(wchar_t) * (len + 1))) == NULL)
	return;
    CopyRegionToWCString(txt, &left, &right, wcs);
#ifdef TRACE
    printf("SetSelectionOwner: len=%d, wstrlen()=%d\n", len, wstrlen(wcs));
#endif
    if ((mbs = (char *)malloc(len * sizeof(MB_LEN_MAX))) != NULL) {
	if ((bytes = wcstombs(mbs, wcs, len * sizeof(MB_LEN_MAX))) >= 0)
	    XStoreBuffer(txt->disp, mbs, bytes, 0);
	free(mbs);
    }
    free(wcs);
#endif
}

void
ResetSelectionOwner(SylTextArea *txt)
{
    XSetSelectionOwner(txt->disp, XA_PRIMARY, None, txt->stamp);
    txt->selected = False;
}

int
IsEqualSylLocation(SylLocation *p, SylLocation *q)
{
    return ((p->x == q->x) && (p->y == q->y));
}

static void
ChangeSelectionOwner(SylTextArea *txt, Time tm)
{
    Window owner;

    txt->selected = !IsEqualSylLocation(&txt->current, &txt->started);
    if ((owner = XGetSelectionOwner(txt->disp, XA_PRIMARY)) == txt->window) {
	if (txt->selected == False)
	    ResetSelectionOwner(txt);
	return;
    }
    if (owner != None) {
        /*
	  Bug?: $B%;%l%/%7%g%s$r=jM-$9$k%&%#%s%I%&$,%H%C%W%l%Y%k%&%#%s%I%&(B
	  $B$G$J$$$?$a$+!"F10l%"%W%j%1!<%7%g%s4V$G4X?t(BXSetSelectionOwner()
	  $B$r8F$S=P$7$F$b(BSelectionClear$B%$%Y%s%H$,$3$J$$$3$H$,$"$k!#BP:v$H(B
	  $B$7$F!"0lEY6/@)E*$K%;%l%/%7%g%s$N%*!<%J$G$J$/$J$k$h$&$K$7$?!#(B

	  $B<+H/E*$KJ|4~$9$k$o$1$G$O$J$$$N$G!"%?%$%`%9%?%s%W$O(Btm$B$G$h$$!#(B
        */
	XSetSelectionOwner(txt->disp, XA_PRIMARY, None, tm);
	/*
	  $B<B$O<!$N$h$&$J%3!<%I$G$b$h$$!#%$%Y%s%H$r(Btxt->window$B$KAw$k$N$O(B
	  $B0l8+$*$+$7$$$h$&$@$1$I!"F10l%"%W%j%1!<%7%g%s$G$O%$%Y%s%H$rEpD0(B
	  $B$7$F$$$k$N$G!"7k2LE*$K$&$^$/$$$/!#(B

	  XEvent ev;
	  
	  ev.xselectionclear.type = SelectionClear;
	  ev.xselectionclear.window = owner;
	  ev.xselectionclear.selection = XA_PRIMARY;
	  ev.xselectionclear.time = tm;
	  XSendEvent(txt->disp, txt->window, False, 0, &ev);
	*/
    }
    SetSelectionOwner(txt, tm);
}

void
DeleteRegion(SylTextArea *txt, SylLocation *left, SylLocation *right)
{
    int y, len, new_x, new_y, width;

    width = TextAreaTextWidth(txt);
    len = SizeOfRegion(txt, left, right);
    InsertLineFeedIntoSylTextBlockSet(txt->tbs, &txt->fontset, width,
				      right->x, right->y, &new_x, &new_y);
    InsertLineFeedIntoSylTextBlockSet(txt->tbs, &txt->fontset, width,
				      left->x, left->y, &new_x, &new_y);
    ++len;
    while (len > 0) {
	for (y = new_y; IsEndOfLineAtLine(txt->tbs, y) == False; ++y)
	    len -= GetLengthAtLine(txt->tbs, y);
	len -= GetLengthAtLine(txt->tbs, y) + 1;
	DeleteToLineEndSylTextBlockSet(txt->tbs, &txt->fontset, width,
				       new_x, new_y, &new_x, &new_y);
	DeleteNextCharInSylTextBlockSet(txt->tbs, &txt->fontset, width,
					new_x, new_y, &new_x, &new_y);
    }
    DeletePrevCharInSylTextBlockSet(txt->tbs, &txt->fontset, width,
				    new_x, new_y, &new_x, &new_y);
    txt->current.x = new_x;
    txt->current.y = new_y;
}

static void
DeleteSelectedRegion(SylTextArea *txt)
{
#if 0
    SylLocation left, right;
    int len, new_x, new_y, width;

    width = TextAreaTextWidth(txt);
    GetSelectedRegion(txt, &left, &right);
    len = SizeOfRegion(txt, &left, &right);
    for (txt->current = right; len > 0; --len) {
	if (txt->current.x > 0) {
	    DeleteLeftCharInSylTextBlockSet(txt->tbs, &txt->fontset, width,
					    txt->current.x, txt->current.y,
					    &new_x, &new_y);
	    txt->current.x = new_x;
	    txt->current.y = new_y;
	}
	else if (txt->current.x == 0 && txt->current.y > 0) {
	    DeletePrevCharInSylTextBlockSet(txt->tbs, &txt->fontset, width,
					    txt->current.x, txt->current.y,
					    &new_x, &new_y);
	    txt->current.x = new_x;
	    txt->current.y = new_y;
	}
    }
#else
    SylLocation left, right;

    GetSelectedRegion(txt, &left, &right);
    DeleteRegion(txt, &left, &right);
#endif
}

static int
GetLimitAtLine(SylTextBlockSet *tbs, int y)
{
    /*
      TextArea2.c$B$K0\9T$7$?$[$&$,$h$$$G$7$g$&!#(B
    */
    int len;

    len = GetLengthAtLine(tbs, y);
    if (IsEndOfLineAtLine(tbs, y) == False && !IsBottomAtLine(tbs, y))
	--len;
    return (len);
}

SylLocation
LocationOfDestination(SylTextArea *txt, SylLocation *origin, int len)
{
    SylLocation goal;
    int x, y, limit;

    x = origin->x;
    y = origin->y;
    for (limit = GetLimitAtLine(txt->tbs, y); len > 0; --len) {
	if (++x > limit) {
	    ++y;
	    limit = GetLimitAtLine(txt->tbs, y);
	    x = 0;
	}
    }
    goal.x = x;
    goal.y = y;
    return (goal);
}

SylLocation
LocationPointed(SylTextArea *txt, int w, int h)
{
    SylLocation p;
    int leftmargin = txt->sidemargin;

    p.y = txt->visible_begin + h / txt->baseline_skip;
    if (p.y < txt->tbs->n_lines) {
	w -= leftmargin;
	p.x = GetRowsAtLine(txt->tbs, &txt->fontset, w, p.y);
	if (p.x == GetLengthAtLine(txt->tbs, p.y)
	    && IsEndOfLineAtLine(txt->tbs, p.y) == False
	    && !IsBottomAtLine(txt->tbs, p.y))
	    --(p.x);
    }
    else {
	p.y = txt->tbs->n_lines - 1;
	p.x = GetLengthAtLine(txt->tbs, p.y);
    }
    return (p);
}

void
BackSpace(SylTextArea *txt, Time tm)
{
    int new_x, new_y, width;

    if (txt->selected) {
	DeleteSelectedRegion(txt);
	ResetSelectionOwner(txt);
    }
    else if (txt->current.x > 0) {
	width = TextAreaTextWidth(txt);
	DeleteLeftCharInSylTextBlockSet(txt->tbs, &txt->fontset, width,
					txt->current.x, txt->current.y,
					&new_x, &new_y);
	txt->current.x = new_x;
	txt->current.y = new_y;
    }
    else if (txt->current.x == 0 && txt->current.y > 0) {
	width = TextAreaTextWidth(txt);
	DeletePrevCharInSylTextBlockSet(txt->tbs, &txt->fontset, width,
					txt->current.x, txt->current.y,
					&new_x, &new_y);
	txt->current.x = new_x;
	txt->current.y = new_y;
    }
    txt->saved_width = GetWidthAtLine(txt->tbs, &txt->fontset,
				      txt->current.x, txt->current.y);
    txt->redraw = True;
}

void
Delete(SylTextArea *t, Time tm)
{
    /*
      $B$A$c$s$H$D$/$C$?$[$&$,$h$$$G$7$g$&!#4X?t(BDeleteRightCharInSylText
      BlockSet()$B$,(BTextArea2.c$B$GI,MW$G$9$M!#(B
    */
    if (t->selected) {
	DeleteSelectedRegion(t);
	ResetSelectionOwner(t);
	t->saved_width = GetWidthAtLine(t->tbs, &t->fontset, t->current.x,
					t->current.y);
	t->redraw = True;
    }
    else {
	if (IsBottomAtLine(t->tbs, t->current.y)
	    && t->current.x >= GetLengthAtLine(t->tbs, t->current.y))
	    return;
	if (++(t->current.x) > GetLimitAtLine(t->tbs, t->current.y)) {
	    ++(t->current.y);
	    t->current.x = 0;
	}
	BackSpace(t, tm);
    }
}

static void
MoveCursorWithSavingWidth(SylTextArea *txt, int x, int y)
{
    txt->current.x = x;
    txt->current.y = y;
    txt->saved_width = GetWidthAtLine(txt->tbs, &txt->fontset, x, y);
    txt->redraw = True;
}

static void
CommonYankLine(SylTextArea *txt)
{
    int x, y, new_x, new_y, width;
    wchar_t *ptr;

    x = txt->current.x;
    y = txt->current.y;
    width = TextAreaTextWidth(txt);
    for (ptr = txt->buf; *ptr != 0; ++ptr) {
	if (*ptr >= ' ' || *ptr == '\t') {
            InsertCharIntoSylTextBlockSet(txt->tbs, &txt->fontset, width,
					  x, y, *ptr, &new_x, &new_y);
        }
	else if (*ptr == '\n') {
            InsertLineFeedIntoSylTextBlockSet(txt->tbs, &txt->fontset, width,
					      x, y, &new_x, &new_y);
	}
	else {
            InsertCharIntoSylTextBlockSet(txt->tbs, &txt->fontset, width,
					  x, y, ' ', &new_x, &new_y);
	}
	x = new_x;
	y = new_y;
    }
    MoveCursorWithSavingWidth(txt, x, y);
}

void
YankLine(SylTextArea *txt, Time tm)
{

    if (txt->buf == NULL)
	return;
    if (txt->selected)
	ResetSelectionOwner(txt);
    CommonYankLine(txt);
}

void
YankLineWithShift(SylTextArea *txt, Time tm)
{
    if (txt->buf == NULL)
	return;
    if (txt->selected == False)
	txt->started = txt->current;
    CommonYankLine(txt);
    ChangeSelectionOwner(txt, tm);
}


static wchar_t *
CreateWCStringOfLineTobeKilled(SylTextArea *txt)
{
    SylLocation head, tail;
    unsigned int len;
    wchar_t *wcs;

    head.x = txt->current.x;
    head.y = txt->current.y;
    for (tail.y = head.y; !IsBottomAtLine(txt->tbs, tail.y)
	     && IsEndOfLineAtLine(txt->tbs, tail.y) == False; ++tail.y)
	;
    tail.x = GetLengthAtLine(txt->tbs, tail.y);

    len = SizeOfRegion(txt, &head, &tail);
    if ((wcs = (wchar_t *)malloc(sizeof(wchar_t) * (len + 1))) == NULL)
	return (NULL);
    CopyRegionToWCString(txt, &head, &tail, wcs);
    return (wcs);
}

static wchar_t *
CreateWCStringOfLineFeed(void)
{
    wchar_t *wcs;

    if ((wcs = (wchar_t *)malloc(sizeof(wchar_t) * 2)) != NULL) {
	wcs[0] = '\n';
	wcs[1] = 0;
    }
    return (wcs);
}

void
KillLine(SylTextArea *txt, Time tm)
{
    int new_x, new_y, width;

    if (txt->selected) {
	/*
	  1998-11-23: $BHO0O;XDj;~$N(BKillLine()$B$,$=$NHO0O$r:o=|$7$F%P%C%U%!$K(B
	  $B%3%T!<$9$k$h$&$K$7$F$_$?!#(B
	*/
	wchar_t *wcs;
	unsigned int len;
	SylLocation left, right;

	GetSelectedRegion(txt, &left, &right);
	len = SizeOfRegion(txt, &left, &right);
	/*
	  CopyRegionToWCString()$B$O:G8e$K%L%kJ8;z$rIU2C$9$k$N$G!"(B1$BJ8;zJ,B?$/(B
	  $B3NJ]$7$F$*$/!#(B
	*/
	if ((wcs = (wchar_t *)malloc(sizeof(wchar_t) * (len + 1))) == NULL)
	    return;
	CopyRegionToWCString(txt, &left, &right, wcs);
	if (txt->buf != NULL)
	    free(txt->buf);
	txt->buf = wcs;
	DeleteSelectedRegion(txt);
	ResetSelectionOwner(txt);
	txt->saved_width = GetWidthAtLine(txt->tbs, &txt->fontset,
					  txt->current.x, txt->current.y);
	txt->redraw = True;
	return;
    }
    width = TextAreaTextWidth(txt);
    if (IsBottomAtLine(txt->tbs, txt->current.y)
	&& txt->current.x >= GetLengthAtLine(txt->tbs, txt->current.y))
	return;
    if (txt->buf != NULL)
	free(txt->buf);
    if (txt->current.x >= GetLengthAtLine(txt->tbs, txt->current.y)
	&& IsEndOfLineAtLine(txt->tbs, txt->current.y) == True) {
	txt->buf = CreateWCStringOfLineFeed();
	DeleteNextCharInSylTextBlockSet(txt->tbs, &txt->fontset, width,
					txt->current.x, txt->current.y,
					&new_x, &new_y);
    }
    else {
	txt->buf = CreateWCStringOfLineTobeKilled(txt);
	DeleteToLineEndSylTextBlockSet(txt->tbs, &txt->fontset, width,
				       txt->current.x, txt->current.y,
				       &new_x, &new_y);
    }
    MoveCursorWithSavingWidth(txt, new_x, new_y);
}

static int
GetHeadLineOfLine(SylTextBlockSet *tbs, int y)
{
    while (y > 0 && IsEndOfLineAtLine(tbs, y - 1) == False)
	--y;
    return (y);
}

static int
GetTailLineOfLine(SylTextBlockSet *tbs, int y)
{
    int n = tbs->n_lines - 1;

    while (y < n && IsEndOfLineAtLine(tbs, y) == False)
	++y;
    return (y);
}

void
CursorBeginingOfLine(SylTextArea *txt, Time tm)
{
    int y;

    y = GetHeadLineOfLine(txt->tbs, txt->current.y);
    if (txt->current.y == y && txt->current.x == 0)
	return;
    if (txt->selected)
	ResetSelectionOwner(txt);
    MoveCursorWithSavingWidth(txt, 0, y);
}

void
CursorBeginingOfLineWithShift(SylTextArea *txt, Time tm)
{
    int y;

    y = GetHeadLineOfLine(txt->tbs, txt->current.y);
    if (txt->current.y == y && txt->current.x == 0)
	return;
    if (txt->selected == False)
	txt->started = txt->current;
    MoveCursorWithSavingWidth(txt, 0, y);
    ChangeSelectionOwner(txt, tm);
}

void
CursorEndOfLine(SylTextArea *txt, Time tm)
{
    int y, s;

    y = GetTailLineOfLine(txt->tbs, txt->current.y);
    s = GetLimitAtLine(txt->tbs, y);
    if (txt->current.y == y && txt->current.x == s)
	return;
    if (txt->selected)
	ResetSelectionOwner(txt);
    MoveCursorWithSavingWidth(txt, s, y);
}

void
CursorEndOfLineWithShift(SylTextArea *txt, Time tm)
{
    int y, s;

    y = GetTailLineOfLine(txt->tbs, txt->current.y);
    s = GetLimitAtLine(txt->tbs, y);
    if (txt->current.y == y && txt->current.x == s)
	return;
    if (txt->selected == False)
	txt->started = txt->current;
    MoveCursorWithSavingWidth(txt, s, y);
    ChangeSelectionOwner(txt, tm);
}

static int
IsSeparator(wchar_t c)
{
    return (ispunct_w(c) || isspace_w(c));
}

static int
IsContinuedForwardDelim(SylTextBlockSet *tbs, int *x, int y)
{
    int n, len;
    SylTextBlock *tb;
    wchar_t *str;

    tb = SeekSylTextBlockAtLine(tbs, y);
    n = y - tbs->current_line;
    len = LengthOfSylText(tb->body[n]);
    str = CreateWCStringFromSylText(tb->body[n], 0, len);
    while (*x < len && IsSeparator(str[*x]))
	++(*x);
    free(str);
    return (*x == len);
}

static int
IsContinuedForwardWord(SylTextBlockSet *tbs, int *x, int y)
{
    int n, len;
    SylTextBlock *tb;
    wchar_t *str;

    tb = SeekSylTextBlockAtLine(tbs, y);
    n = y - tbs->current_line;
    len = LengthOfSylText(tb->body[n]);
    str = CreateWCStringFromSylText(tb->body[n], 0, len);
    while (*x < len && !IsSeparator(str[*x]))
	++(*x);
    free(str);
    return (*x == len && IsEndOfLineAtLine(tbs, y) == False);
}

static void
CommonCursorNextWord(SylTextArea *txt, int x, int y)
{
    SylTextBlockSet *tbs = txt->tbs;

    while (IsContinuedForwardDelim(tbs, &x, y) && !IsBottomAtLine(tbs, y)) {
	++y;
	x = 0;
    }
    while (IsContinuedForwardWord(tbs, &x, y) && !IsBottomAtLine(tbs, y)) {
	++y;
	x = 0;
    }
    if (x == GetLengthAtLine(tbs, y) && IsEndOfLineAtLine(tbs, y) == False
	&& IsBottomAtLine(tbs, y) == False) {
	/* NEVER ($B%"%k%4%j%:%`E*$K$"$j$($J$$(B) */
	++y;
	x = 0;
    }
    MoveCursorWithSavingWidth(txt, x, y);
}

void
CursorNextWord(SylTextArea *txt, Time tm)
{
    if (IsBottomAtLine(txt->tbs, txt->current.y)
	&& txt->current.x >= GetLengthAtLine(txt->tbs, txt->current.y))
	return; 
    if (txt->selected)
	ResetSelectionOwner(txt);
    CommonCursorNextWord(txt, txt->current.x, txt->current.y);
}

void
CursorNextWordWithShift(SylTextArea *txt, Time tm)
{
    if (IsBottomAtLine(txt->tbs, txt->current.y)
	&& txt->current.x >= GetLengthAtLine(txt->tbs, txt->current.y))
	return; 
    if (txt->selected == False)
	txt->started = txt->current;
    CommonCursorNextWord(txt, txt->current.x, txt->current.y);
    ChangeSelectionOwner(txt, tm);
}

static int
IsContinuedBackwardDelim(SylTextBlockSet *tbs, int *x, int y)
{
    int n, len;
    SylTextBlock *tb;
    wchar_t *str;

    tb = SeekSylTextBlockAtLine(tbs, y);
    n = y - tbs->current_line;
    len = LengthOfSylText(tb->body[n]);
    str = CreateWCStringFromSylText(tb->body[n], 0, len);
    while (*x > 0 && IsSeparator(str[*x - 1]))
	--(*x);
    free(str);
    return (*x == 0);
}

static int
IsContinuedBackwardWord(SylTextBlockSet *tbs, int *x, int y)
{
    int n, len;
    SylTextBlock *tb;
    wchar_t *str;

    tb = SeekSylTextBlockAtLine(tbs, y);
    n = y - tbs->current_line;
    len = LengthOfSylText(tb->body[n]);
    str = CreateWCStringFromSylText(tb->body[n], 0, len);
    while (*x > 0 && !IsSeparator(str[*x - 1]))
	--(*x);
    free(str);
    return (*x == 0 && y > 0 && IsEndOfLineAtLine(tbs, y - 1) == False);
}

static void
CommonCursorPrevWord(SylTextArea *txt, int x, int y)
{
    SylTextBlockSet *tbs = txt->tbs;

    while (IsContinuedBackwardDelim(tbs, &x, y) && y > 0) {
	--y;
	x = GetLengthAtLine(tbs, y);
    }
    while (IsContinuedBackwardWord(tbs, &x, y) && y > 0) {
	--y;
	x = GetLengthAtLine(tbs, y);
    }
    if (x == GetLengthAtLine(tbs, y) && IsEndOfLineAtLine(tbs, y) == False
	&& IsBottomAtLine(tbs, y) == False) {
	++y;
	x = 0;
    }
    MoveCursorWithSavingWidth(txt, x, y);
}

void
CursorPrevWord(SylTextArea *txt, Time tm)
{
    if (txt->current.y <= 0 && txt->current.x <= 0)
	return; 
    if (txt->selected)
	ResetSelectionOwner(txt);
    CommonCursorPrevWord(txt, txt->current.x, txt->current.y);
}

void
CursorPrevWordWithShift(SylTextArea *txt, Time tm)
{
    if (txt->current.y <= 0 && txt->current.x <= 0)
	return; 
    if (txt->selected == False)
	txt->started = txt->current;
    CommonCursorPrevWord(txt, txt->current.x, txt->current.y);
    ChangeSelectionOwner(txt, tm);
}

static void
CommonCursorLeft(SylTextArea *txt)
{
    if (--(txt->current.x) < 0) {
	--(txt->current.y);
	txt->current.x = GetLimitAtLine(txt->tbs, txt->current.y);
    }
    txt->saved_width = GetWidthAtLine(txt->tbs, &txt->fontset,
				      txt->current.x, txt->current.y);
    txt->redraw = True;
}

void
CursorLeft(SylTextArea *txt, Time tm)
{
    if (txt->current.y <= 0 && txt->current.x <= 0)
	return;
    if (txt->selected)
	ResetSelectionOwner(txt);
    CommonCursorLeft(txt);
}

void
CursorLeftWithShift(SylTextArea *txt, Time tm)
{
    if (txt->current.y <= 0 && txt->current.x <= 0)
	return;
    if (txt->selected == False)
	txt->started = txt->current;
    CommonCursorLeft(txt);
    ChangeSelectionOwner(txt, tm);
}

static void
CommonCursorRight(SylTextArea *txt)
{
    if (++(txt->current.x) > GetLimitAtLine(txt->tbs, txt->current.y)) {
	++(txt->current.y);
	txt->current.x = 0;
    }
    txt->saved_width = GetWidthAtLine(txt->tbs, &txt->fontset,
				      txt->current.x, txt->current.y);
    txt->redraw = True;
}

void
CursorRight(SylTextArea *txt, Time tm)
{
    if (IsBottomAtLine(txt->tbs, txt->current.y)
	&& txt->current.x >= GetLengthAtLine(txt->tbs, txt->current.y))
	return; 
    if (txt->selected)
	ResetSelectionOwner(txt);
    CommonCursorRight(txt);
}

void
CursorRightWithShift(SylTextArea *txt, Time tm)
{
    if (IsBottomAtLine(txt->tbs, txt->current.y)
	&& txt->current.x >= GetLengthAtLine(txt->tbs, txt->current.y))
	return; 
    if (txt->selected == False)
	txt->started = txt->current;
    CommonCursorRight(txt);
    ChangeSelectionOwner(txt, tm);
}

void
InsertChar(SylTextArea *txt, wchar_t c)
{
    int new_x, new_y, width;

    if (txt->selected) {
	DeleteSelectedRegion(txt);
	ResetSelectionOwner(txt);
    }
    width = TextAreaTextWidth(txt);
    InsertCharIntoSylTextBlockSet(txt->tbs, &txt->fontset, width,
				  txt->current.x, txt->current.y, c,
				  &new_x, &new_y);
    MoveCursorWithSavingWidth(txt, new_x, new_y);
}

void
LineFeed(SylTextArea *txt, Time tm)
{
    int new_x, new_y, width;

    if (txt->selected) {
	DeleteSelectedRegion(txt);
	ResetSelectionOwner(txt);
    }
    width = TextAreaTextWidth(txt);
    InsertLineFeedIntoSylTextBlockSet(txt->tbs, &txt->fontset, width,
				      txt->current.x, txt->current.y,
				      &new_x, &new_y);
    MoveCursorWithSavingWidth(txt, new_x, new_y);
}

static void
CommonEdge(SylTextArea *txt, int x)
{
    txt->current.x = x;
    txt->saved_width = GetWidthAtLine(txt->tbs, &txt->fontset,
				      txt->current.x, txt->current.y);
    txt->redraw = True;
}

void
CursorLeftEdge(SylTextArea *txt, Time tm)
{
    if (txt->current.x > 0) {
	if (txt->selected)
	    ResetSelectionOwner(txt);
	CommonEdge(txt, 0);
    }    
}

void
CursorLeftEdgeWithShift(SylTextArea *txt, Time tm)
{
    if (txt->current.x > 0) {
	if (txt->selected == False)
	    txt->started = txt->current;
	CommonEdge(txt, 0);
	ChangeSelectionOwner(txt, tm);
    }    
}

void
CursorRightEdge(SylTextArea *txt, Time tm)
{
    int s = GetLimitAtLine(txt->tbs, txt->current.y);

    if (txt->current.x < s) {
	if (txt->selected)
	    ResetSelectionOwner(txt);
	CommonEdge(txt, s);
    }    
}

void
CursorRightEdgeWithShift(SylTextArea *txt, Time tm)
{
    int s = GetLimitAtLine(txt->tbs, txt->current.y);

    if (txt->current.x < s) {
	if (txt->selected == False)
	    txt->started = txt->current;
	CommonEdge(txt, s);
	ChangeSelectionOwner(txt, tm);
    }    
}

int
RowsAfterMovingVertically(SylTextArea *txt)
{
    int x;

    x = GetRowsAtLine(txt->tbs, &txt->fontset, txt->saved_width,
		      txt->current.y);
    if (x == GetLengthAtLine(txt->tbs, txt->current.y)
	&& IsEndOfLineAtLine(txt->tbs, txt->current.y) == False
	&& !IsBottomAtLine(txt->tbs, txt->current.y))
	--x;
    return (x);
}

static void
CommonCursorUp(SylTextArea *txt)
{
    --(txt->current.y);
    txt->current.x = RowsAfterMovingVertically(txt);
    txt->redraw = True;
}

void
CursorUp(SylTextArea *txt, Time tm)
{
    if (txt->current.y > 0) {
	if (txt->selected)
	    ResetSelectionOwner(txt);
	CommonCursorUp(txt);
    }
}

void
CursorUpWithShift(SylTextArea *txt, Time tm)
{
    if (txt->current.y > 0) {
	if (txt->selected == False)
	    txt->started = txt->current;
	CommonCursorUp(txt);
	ChangeSelectionOwner(txt, tm);
    }
}

static void
CommonCursorDown(SylTextArea *txt)
{
    ++(txt->current.y);
    txt->current.x = RowsAfterMovingVertically(txt);
    txt->redraw = True;
}

void
CursorDown(SylTextArea *txt, Time tm)
{
    if (!IsBottomAtLine(txt->tbs, txt->current.y)) {
	if (txt->selected)
	    ResetSelectionOwner(txt);
	CommonCursorDown(txt);
    }
}

void
CursorDownWithShift(SylTextArea *txt, Time tm)
{
    if (!IsBottomAtLine(txt->tbs, txt->current.y)) {
	if (txt->selected == False)
	    txt->started = txt->current;
	CommonCursorDown(txt);
	ChangeSelectionOwner(txt, tm);
    }
}

void
GrabbedCursorUp(SylTextArea *txt)
{
    if (txt->current.y > 0)
	CommonCursorUp(txt);
}

void
GrabbedCursorDown(SylTextArea *txt)
{
    if (!IsBottomAtLine(txt->tbs, txt->current.y))
	CommonCursorDown(txt);
}

void
SetFocusToNextField(SylTextArea *txt, Time tm)
{
    if (txt->next_field != None)
	XSetInputFocus(txt->disp, txt->next_field, RevertToParent, tm);
}

void
SetFocusToPrevField(SylTextArea *txt, Time tm)
{
    if (txt->prev_field != None)
	XSetInputFocus(txt->disp, txt->prev_field, RevertToParent, tm);
}

static void
CommonScrollUp(SylTextArea *txt)
{
    txt->visible_begin = max(0, txt->visible_begin - txt->visible_cols);
    txt->current.y = max(0, txt->current.y - txt->visible_cols);
    txt->current.x = RowsAfterMovingVertically(txt);
    txt->redraw = True;
}

void
ScrollUp(SylTextArea *txt, Time tm)
{
    if (txt->current.y > 0) {
	if (txt->selected)
	    ResetSelectionOwner(txt);
	CommonScrollUp(txt);
    }
}

void
ScrollUpWithShift(SylTextArea *txt, Time tm)
{
    if (txt->current.y > 0) {
	if (txt->selected == False)
	    txt->started = txt->current;
	CommonScrollUp(txt);
	ChangeSelectionOwner(txt, tm);
    }
}

static void
CommonScrollDown(SylTextArea *txt)
{
    txt->visible_begin = max(0, min(txt->tbs->n_lines - txt->visible_cols,
				    txt->visible_begin + txt->visible_cols));
    txt->current.y = min(txt->tbs->n_lines - 1,
			 txt->current.y + txt->visible_cols);
    txt->current.x = RowsAfterMovingVertically(txt);
    txt->redraw = True;
}

void
ScrollDown(SylTextArea *txt, Time tm)
{
    if (!IsBottomAtLine(txt->tbs, txt->current.y)) {
	if (txt->selected)
	    ResetSelectionOwner(txt);
	CommonScrollDown(txt);
    }
}

void
ScrollDownWithShift(SylTextArea *txt, Time tm)
{
    if (!IsBottomAtLine(txt->tbs, txt->current.y)) {
	if (txt->selected == False)
	    txt->started = txt->current;
	CommonScrollDown(txt);
	ChangeSelectionOwner(txt, tm);
    }
}
